-- © Copyright 2021 Pedro Gimeno Fortea. All rights reserved.

local size = 128
local img = love.image.newImageData(size+1, size+1, "r8", nil)
local ffi = require 'ffi'
local ptr = ffi.cast('unsigned char *', img:getFFIPointer())
local quad = love.graphics.newQuad(0, 0, size, size, size+1, size+1)

local gen = love.math.newRandomGenerator()

local map = {}
local mapX = 64
local mapY = 64
local speed = 200

local function scaleClamp(x)
  return math.max(0, math.min(255, 255.00000000001 * x))
end

local byte2float = 1.00000000000001 / 255
local levelFactor = math.sqrt(0.5)

local function diamondSquare(x1, y1, x2, y2, level)
  local scale = size
  while scale > 1 do
    scale = scale * 0.5
    level = level * levelFactor
    -- average/fill in X
    for y = scale, size - scale, scale+scale do
      for x = scale, size - scale, scale+scale do
        local avg = ptr[(x - scale) + (y - scale)*(size+1)] * byte2float
        avg = avg + ptr[(x + scale) + (y - scale)*(size+1)] * byte2float
        avg = avg + ptr[(x - scale) + (y + scale)*(size+1)] * byte2float
        avg = (avg + ptr[(x + scale) + (y + scale)*(size+1)] * byte2float)*0.25
        ptr[x + y*(size+1)] = scaleClamp(avg + (gen:random()-0.5)*level)
      end
    end
    -- average/fill in +, part 1 (top pixels)
    for y = scale+scale, size - scale, scale+scale do
      for x = scale, size - scale, scale+scale do
        local avg = ptr[(x - scale) + y*(size+1)] * byte2float
        avg = avg + ptr[(x + scale) + y*(size+1)] * byte2float
        avg = avg + ptr[x + (y - scale)*(size+1)] * byte2float
        avg = (avg + ptr[x + (y + scale)*(size+1)] * byte2float) * 0.25
        ptr[x + y*(size+1)] = scaleClamp(avg + (gen:random()-0.5)*level)
      end
    end
    -- average/fill in +, part 2 (left pixels)
    for y = scale, size - scale, scale+scale do
      for x = scale+scale, size - scale, scale+scale do
        local avg = ptr[(x - scale) + y*(size+1)] * byte2float
        avg = avg + ptr[(x + scale) + y*(size+1)] * byte2float
        avg = avg + ptr[x + (y - scale)*(size+1)] * byte2float
        avg = (avg + ptr[x + (y + scale)*(size+1)] * byte2float) * 0.25
        ptr[x + y*(size+1)] = scaleClamp(avg + (gen:random()-0.5)*level)
      end
    end
  end
end

local function makeHBorder(x1, x2, y, level)
  local mid = (x1 + x2) * 0.5
  local prevPixel = ptr[x1 + y*(size+1)] * byte2float
  local nextPixel = ptr[x2 + y*(size+1)] * byte2float
  ptr[mid + y*(size+1)] = scaleClamp((prevPixel + nextPixel)*0.5 + (gen:random() * 2 - 1) * level)
  if mid + 1 < x2 then
    makeHBorder(x1, mid, y, level*levelFactor^2)
    makeHBorder(mid, x2, y, level*levelFactor^2)
  end
end

local function makeVBorder(y1, y2, x, level)
  local mid = (y1 + y2) * 0.5
  local prevPixel = ptr[x + y1*(size+1)] * byte2float
  local nextPixel = ptr[x + y2*(size+1)] * byte2float
  ptr[x + mid*(size+1)] = scaleClamp((prevPixel + nextPixel)*0.5 + (gen:random()*2 - 1) * level)
  assert(mid % 1 == 0 and y1 % 1 == 0 and y2 % 1 == 0)
  if mid + 1 < y2 then
    makeVBorder(y1, mid, x, level*levelFactor^2)
    makeVBorder(mid, y2, x, level*levelFactor^2)
  end
end

local function generateImage(gx, gy)
  if not map[gy] then
    map[gy] = {}
  end

  local level = 0.5

  -- Top left corner
  gen:setSeed(gx, gy)
  ptr[0] = scaleClamp(gen:random() * 0.5 + 0.25)

  -- Bottom right corner
  gen:setSeed(gx+1, gy+1)
  ptr[size + size*(size+1)] = scaleClamp(gen:random() * 0.5 + 0.25)

  -- Bottom left corner
  gen:setSeed(gx, gy+1)
  ptr[size * (size+1)] = scaleClamp(gen:random() * 0.5 + 0.25)
  -- We can create the bottom border now = top border of bottom square
  makeHBorder(0, size, size, level)

  -- Top right corner
  gen:setSeed(gx+1, gy)
  ptr[size] = scaleClamp(gen:random() * 0.5 + 0.25)
  -- Advance the RNG as if the left border was generated
  for i = 2, size do gen:random() end
  -- Create right border = left border of right square
  makeVBorder(0, size, size, level)

  gen:setSeed(gx, gy)
  -- Advance the generator by the top left pixel
  gen:random()
  -- Make top border
  makeHBorder(0, size, 0, level)
  -- Make left border
  makeVBorder(0, size, 0, level)

  -- Finally, fill in the interior
  diamondSquare(0, 0, size, size, level)
  map[gy][gx] = love.graphics.newImage(img)
end

local pressed = love.keyboard.isScancodeDown

local mask = love.image.newImageData(6, 6, "rgba8",
  "\0\0\0\255\0\0\0\255\0\0\0\255\0\0\0\255\0\0\0\255\0\0\0\255" ..
  "\0\0\0\255\0\0\0\255\0\0\0\255\0\0\0\255\0\0\0\255\0\0\0\255" ..
  "\0\0\0\255\0\0\0\255\0\0\0\000\0\0\0\000\0\0\0\255\0\0\0\255" ..
  "\0\0\0\255\0\0\0\255\0\0\0\000\0\0\0\000\0\0\0\255\0\0\0\255" ..
  "\0\0\0\255\0\0\0\255\0\0\0\255\0\0\0\255\0\0\0\255\0\0\0\255" ..
  "\0\0\0\255\0\0\0\255\0\0\0\255\0\0\0\255\0\0\0\255\0\0\0\255")
local maskImg = love.graphics.newImage(mask)
maskImg:setFilter("nearest", "nearest")

function love.update(dt)
  if pressed("up", "w") then
    mapY = mapY - speed * dt
  end
  if pressed("down", "s") then
    mapY = mapY + speed * dt
  end
  if pressed("left", "a") then
    mapX = mapX - speed * dt
  end
  if pressed("right", "d") then
    mapX = mapX + speed * dt
  end
  local gridX = math.floor(mapX/size)
  local gridY = math.floor(mapY/size)
  for y = gridY - 1, gridY + 1 do
    if not map[y] then
      map[y] = {}
    end
    for x = gridX - 1, gridX + 1 do
      if not map[y][x] then
        generateImage(x, y)
      end
    end
  end
  -- destroy the images at a distance of 2 to avoid filling the VRAM
  -- (this should be done with a distance sufficient for them to go out
  -- of the screen though)
  for y = gridY - 2, gridY + 2 do
    if map[y] then
      if map[y][gridX - 2] then
        map[y][gridX - 2]:release()
        map[y][gridX - 2] = nil
      end
      if map[y][gridX + 2] then
        map[y][gridX + 2]:release()
        map[y][gridX + 2] = nil
      end
    end
  end
  if map[gridY - 2] then
    for x = gridX - 1, gridX + 1 do
      if map[gridY - 2][x] then
        map[gridY - 2][x]:release()
        map[gridY - 2][x] = nil
      end
    end
  end
  if map[gridY + 2] then
    for x = gridX - 1, gridX + 1 do
      if map[gridY + 2][x] then
        map[gridY + 2][x]:release()
        map[gridY + 2][x] = nil
      end
    end
  end
-- in my computer, enabling this line causes an FPS drop from 820 to 710,
-- which means it takes about 1.4 ms to execute
--  generateImage(1, 1)
  love.window.setTitle("FPS: " .. love.timer.getFPS())
end

-- Draw in greyscale instead of red
local shader = love.graphics.newShader[[
  vec4 effect(vec4 colour, Image tex, vec2 texpos, vec2 scrpos)
  {
    return Texel(tex, texpos).rrra * colour;
  }
]]

function love.draw()
  local hw, hh = love.graphics.getDimensions()
  hw, hh = hw*0.5, hh*0.5

  local gridX = math.floor(mapX/size)
  local gridY = math.floor(mapY/size)
  love.graphics.setShader(shader)
  for y = gridY - 1, gridY + 1 do
    if map[y] then
      for x = gridX - 1, gridX + 1 do
        if map[y][x] then
          love.graphics.draw(map[y][x], quad, x*size-mapX + hw, y*size-mapY + hh)
        end
      end
    end
  end
  love.graphics.setShader()
  love.graphics.setColor(1, 1, 1, 0.85)
  love.graphics.draw(maskImg, hw, hh, 0, 128, 128, 3, 3)
  love.graphics.setColor(1, 1, 1)
end

function love.keypressed(k) return k == "escape" and love.event.quit() end
